@@ -5,8 +5,9 @@ extern crate tao;
 
 use simple_logger::SimpleLogger;
 use tao::{
-  event::{Event, VirtualKeyCode, WindowEvent},
+  event::{Event, WindowEvent},
   event_loop::{ControlFlow, EventLoop},
+  keyboard::Key,
   window::WindowBuilder,
 };
 
@@ -31,15 +32,13 @@ fn main() {
 
       // Keyboard input event to handle minimize via a hotkey
       Event::WindowEvent {
-        event: WindowEvent::KeyboardInput { input, .. },
+        event: WindowEvent::KeyboardInput { event, .. },
         window_id,
-      } => {
-        if window_id == window.id() {
-          // Pressing the 'M' key will minimize the window
-          if input.virtual_keycode == Some(VirtualKeyCode::M) {
-            window.set_minimized(true);
-          }
-        }
+      } if window_id == window.id() && Key::Character("m") == event.logical_key => {
+        // Pressing the 'm' key will minimize the window
+        // WARNING: Consider using `key_without_modifers()` if available on your platform.
+        // See the `key_binding` example
+        window.set_minimized(true);
       }
       _ => (),
     }
