@@ -7,7 +7,8 @@ use std::{
   error::Error,
   process,
   rc::Rc,
-  sync::mpsc::SendError,
+  sync::{mpsc::SendError, Mutex},
+  time::Instant,
 };
 
 use gdk::{Cursor, CursorType, WindowExt, WindowState};
@@ -21,19 +22,18 @@ use glib::Cast;
 use gtk::{Clipboard, Entry};
 
 use crate::{
+  accelerator::AcceleratorId,
   dpi::{PhysicalPosition, PhysicalSize},
-  event::{
-    DeviceId as RootDeviceId, ElementState, Event, ModifiersState, MouseButton, StartCause,
-    WindowEvent,
-  },
+  event::{DeviceId as RootDeviceId, ElementState, Event, MouseButton, StartCause, WindowEvent},
   event_loop::{ControlFlow, EventLoopClosed, EventLoopWindowTarget as RootELW},
+  keyboard::ModifiersState,
   menu::{MenuItem, MenuType},
   monitor::MonitorHandle as RootMonitorHandle,
   window::{CursorIcon, WindowId as RootWindowId},
 };
 
 use super::{
-  menu,
+  keyboard,
   monitor::MonitorHandle,
   window::{WindowId, WindowRequest},
   DeviceId,
@@ -137,7 +137,6 @@ impl<T: 'static> EventLoop<T> {
 
     let context = MainContext::default();
     context.push_thread_default();
-    let keep_running = Rc::new(RefCell::new(true));
 
     // User event
     let event_tx_ = event_tx.clone();
@@ -152,7 +151,6 @@ impl<T: 'static> EventLoop<T> {
     let app = window_target.p.app.clone();
     let windows = window_target.p.windows.clone();
     let window_requests_tx = window_target.p.window_requests_tx.clone();
-    let keep_running_ = keep_running.clone();
     self
       .window_requests_rx
       .attach(Some(&context), move |(id, request)| {
@@ -249,7 +247,7 @@ impl<T: 'static> EventLoop<T> {
                 window.set_urgency_hint(true)
               }
             }
-            WindowRequest::SkipTaskbar => window.set_skip_taskbar_hint(true),
+            WindowRequest::SetSkipTaskbar(skip) => window.set_skip_taskbar_hint(skip),
             WindowRequest::CursorIcon(cursor) => {
               if let Some(gdk_window) = window.get_window() {
                 let display = window.get_display();
@@ -482,134 +480,282 @@ impl<T: 'static> EventLoop<T> {
               });
             }
             WindowRequest::Redraw => window.queue_draw(),
-            WindowRequest::Menu(m) => {
-              match m {
-                MenuItem::Custom(c) => {
-                  if let Err(e) = event_tx.send(Event::MenuEvent {
-                    menu_id: c.id,
-                    origin: MenuType::Menubar,
-                  }) {
-                    log::warn!("Failed to send menu event to event channel: {}", e);
-                  }
+            WindowRequest::Menu(m) => match m {
+              (None, Some(menu_id)) => {
+                if let Err(e) = event_tx.send(Event::MenuEvent {
+                  menu_id,
+                  origin: MenuType::MenuBar,
+                }) {
+                  log::warn!("Failed to send menu event to event channel: {}", e);
                 }
-                MenuItem::About(_) => {
-                  let about = AboutDialog::new();
-                  about.show_all();
-                  app.add_window(&about);
+              }
+              (Some(MenuItem::About(_)), None) => {
+                let about = AboutDialog::new();
+                about.show_all();
+                app.add_window(&about);
+              }
+              (Some(MenuItem::Hide), None) => window.hide(),
+              (Some(MenuItem::CloseWindow), None) => window.close(),
+              (Some(MenuItem::Quit), None) => {
+                if let Err(e) = event_tx.send(Event::LoopDestroyed) {
+                  log::warn!(
+                    "Failed to send loop destroyed event to event channel: {}",
+                    e
+                  );
                 }
-                MenuItem::Hide => window.hide(),
-                MenuItem::CloseWindow => window.close(),
-                MenuItem::Quit => {
-                  keep_running_.replace(false);
-                }
-                #[cfg(any(feature = "menu", feature = "tray"))]
-                MenuItem::Cut => {
-                  if let Some(widget) = window.get_focus() {
-                    if widget.has_focus() {
-                      if let Some(view) = widget.dynamic_cast_ref::<sourceview::View>() {
-                        if let Some(clipboard) = Clipboard::get_default(&widget.get_display()) {
-                          if let Some(buf) = view.get_buffer() {
-                            buf.cut_clipboard(&clipboard, true);
-                          }
-                        }
-                      } else if let Some(entry) = widget.dynamic_cast_ref::<Entry>() {
-                        entry.cut_clipboard();
-                      }
-                    }
-                  }
-                }
-                #[cfg(any(feature = "menu", feature = "tray"))]
-                MenuItem::Copy => {
-                  if let Some(widget) = window.get_focus() {
-                    if widget.has_focus() {
-                      if let Some(view) = widget.dynamic_cast_ref::<sourceview::View>() {
-                        if let Some(clipboard) = Clipboard::get_default(&widget.get_display()) {
-                          if let Some(buf) = view.get_buffer() {
-                            buf.copy_clipboard(&clipboard);
-                          }
-                        }
-                      } else if let Some(entry) = widget.dynamic_cast_ref::<Entry>() {
-                        entry.copy_clipboard();
-                      }
-                    }
-                  }
-                }
-                #[cfg(any(feature = "menu", feature = "tray"))]
-                MenuItem::Paste => {
-                  if let Some(widget) = window.get_focus() {
-                    if widget.has_focus() {
-                      if let Some(view) = widget.dynamic_cast_ref::<sourceview::View>() {
-                        if let Some(clipboard) = Clipboard::get_default(&widget.get_display()) {
-                          if let Some(buf) = view.get_buffer() {
-                            buf.paste_clipboard(&clipboard, None, true);
-                          }
-                        }
-                      } else if let Some(entry) = widget.dynamic_cast_ref::<Entry>() {
-                        entry.paste_clipboard();
-                      }
-                    }
-                  }
-                }
-                #[cfg(any(feature = "menu", feature = "tray"))]
-                MenuItem::SelectAll => {
-                  if let Some(widget) = window.get_focus() {
-                    if widget.has_focus() {
-                      if let Some(view) = widget.dynamic_cast_ref::<sourceview::View>() {
+              }
+              #[cfg(any(feature = "menu", feature = "tray"))]
+              (Some(MenuItem::Cut), None) => {
+                if let Some(widget) = window.get_focus() {
+                  if widget.has_focus() {
+                    if let Some(view) = widget.dynamic_cast_ref::<sourceview::View>() {
+                      if let Some(clipboard) = Clipboard::get_default(&widget.get_display()) {
                         if let Some(buf) = view.get_buffer() {
-                          buf.select_range(&buf.get_start_iter(), &buf.get_end_iter());
+                          buf.cut_clipboard(&clipboard, true);
                         }
-                      } else if let Some(entry) = widget.dynamic_cast_ref::<Entry>() {
-                        entry.select_region(0, -1);
                       }
+                    } else if let Some(entry) = widget.dynamic_cast_ref::<Entry>() {
+                      entry.cut_clipboard();
                     }
                   }
                 }
-                // TODO toggle fullscreen
-                MenuItem::EnterFullScreen => window.fullscreen(),
-                MenuItem::Minimize => window.iconify(),
-                _ => {}
+              }
+              #[cfg(any(feature = "menu", feature = "tray"))]
+              (Some(MenuItem::Copy), None) => {
+                if let Some(widget) = window.get_focus() {
+                  if widget.has_focus() {
+                    if let Some(view) = widget.dynamic_cast_ref::<sourceview::View>() {
+                      if let Some(clipboard) = Clipboard::get_default(&widget.get_display()) {
+                        if let Some(buf) = view.get_buffer() {
+                          buf.copy_clipboard(&clipboard);
+                        }
+                      }
+                    } else if let Some(entry) = widget.dynamic_cast_ref::<Entry>() {
+                      entry.copy_clipboard();
+                    }
+                  }
+                }
+              }
+              #[cfg(any(feature = "menu", feature = "tray"))]
+              (Some(MenuItem::Paste), None) => {
+                if let Some(widget) = window.get_focus() {
+                  if widget.has_focus() {
+                    if let Some(view) = widget.dynamic_cast_ref::<sourceview::View>() {
+                      if let Some(clipboard) = Clipboard::get_default(&widget.get_display()) {
+                        if let Some(buf) = view.get_buffer() {
+                          buf.paste_clipboard(&clipboard, None, true);
+                        }
+                      }
+                    } else if let Some(entry) = widget.dynamic_cast_ref::<Entry>() {
+                      entry.paste_clipboard();
+                    }
+                  }
+                }
+              }
+              #[cfg(any(feature = "menu", feature = "tray"))]
+              (Some(MenuItem::SelectAll), None) => {
+                if let Some(widget) = window.get_focus() {
+                  if widget.has_focus() {
+                    if let Some(view) = widget.dynamic_cast_ref::<sourceview::View>() {
+                      if let Some(buf) = view.get_buffer() {
+                        buf.select_range(&buf.get_start_iter(), &buf.get_end_iter());
+                      }
+                    } else if let Some(entry) = widget.dynamic_cast_ref::<Entry>() {
+                      entry.select_region(0, -1);
+                    }
+                  }
+                }
+              }
+              (Some(MenuItem::EnterFullScreen), None) => {
+                let state = window.get_window().unwrap().get_state();
+                if state.contains(WindowState::FULLSCREEN) {
+                  window.unfullscreen();
+                } else {
+                  window.fullscreen();
+                }
+              }
+              (Some(MenuItem::Minimize), None) => window.iconify(),
+              _ => {}
+            },
+            WindowRequest::SetMenu((window_menu, accel_group, mut menubar)) => {
+              if let Some(window_menu) = window_menu {
+                // remove all existing elements as we overwrite
+                // but we keep same menubar reference
+                for i in menubar.get_children() {
+                  menubar.remove(&i);
+                }
+                // create all new elements
+                window_menu.generate_menu(&mut menubar, &window_requests_tx, &accel_group, id);
+                // make sure all newly added elements are visible
+                menubar.show_all();
               }
             }
-            WindowRequest::SetMenu((menus, accel_group, menu)) => {
-              for i in menu.get_children() {
-                menu.remove(&i);
+            WindowRequest::KeyboardInput((event_key, element_state)) => {
+              // if we have a modifier lets send it
+              let mut mods = keyboard::get_modifiers(event_key.clone());
+              if !mods.is_empty() {
+                // if we release the modifier tell the world
+                if ElementState::Released == element_state {
+                  mods = ModifiersState::empty();
+                }
+
+                if let Err(e) = event_tx.send(Event::WindowEvent {
+                  window_id: RootWindowId(id),
+                  event: WindowEvent::ModifiersChanged(mods),
+                }) {
+                  log::warn!(
+                    "Failed to send modifiers changed event to event channel: {}",
+                    e
+                  );
+                } else {
+                  // stop here we don't want to send the key event
+                  // as we emit the `ModifiersChanged`
+                  return Continue(true);
+                }
               }
 
-              if let Some(menus) = menus {
-                let menubar = menu::initialize(id, menus, &window_requests_tx, &accel_group);
-                menu.pack_start(&menubar, false, false, 0);
-                menu.show_all();
+              // todo: implement repeat?
+              let event = keyboard::make_key_event(&event_key, false, None, element_state);
+
+              if let Some(event) = event {
+                if let Err(e) = event_tx.send(Event::WindowEvent {
+                  window_id: RootWindowId(id),
+                  event: WindowEvent::KeyboardInput {
+                    // FIXME: currently we use a dummy device id, find if we can get device id from gtk
+                    device_id: RootDeviceId(DeviceId(0)),
+                    event,
+                    is_synthetic: false,
+                  },
+                }) {
+                  log::warn!("Failed to send keyboard event to event channel: {}", e);
+                }
               }
             }
+            // we use dummy window id for `GlobalHotKey`
+            WindowRequest::GlobalHotKey(_hotkey_id) => {}
           }
         } else if id == WindowId::dummy() {
-          if let WindowRequest::Menu(MenuItem::Custom(c)) = request {
-            if let Err(e) = event_tx.send(Event::MenuEvent {
-              menu_id: c.id,
-              origin: MenuType::SystemTray,
-            }) {
-              log::warn!("Failed to send status bar event to event channel: {}", e);
+          match request {
+            WindowRequest::GlobalHotKey(hotkey_id) => {
+              if let Err(e) = event_tx.send(Event::GlobalShortcutEvent(AcceleratorId(hotkey_id))) {
+                log::warn!("Failed to send global hotkey event to event channel: {}", e);
+              }
             }
+            WindowRequest::Menu((None, Some(menu_id))) => {
+              if let Err(e) = event_tx.send(Event::MenuEvent {
+                menu_id,
+                origin: MenuType::ContextMenu,
+              }) {
+                log::warn!("Failed to send status bar event to event channel: {}", e);
+              }
+            }
+            _ => {}
           }
         }
         Continue(true)
       });
 
     // Event control flow
-    let keep_running_ = keep_running.clone();
+    let events = Rc::new(Mutex::new(Vec::new()));
+    let events_ = events.clone();
     event_rx.attach(Some(&context), move |event| {
-      callback(event, &window_target, &mut control_flow);
-
-      match control_flow {
-        ControlFlow::Exit => {
-          keep_running_.replace(false);
-        }
-        _ => {}
-      }
+      let mut e = events_.lock().unwrap();
+      e.push(event);
       Continue(true)
     });
 
-    while *keep_running.borrow() {
+    loop {
+      match control_flow {
+        ControlFlow::Exit => {
+          callback(Event::LoopDestroyed, &window_target, &mut control_flow);
+          break;
+        }
+        ControlFlow::Wait => {
+          let mut e = events.lock().unwrap();
+          if !e.is_empty() {
+            callback(
+              Event::NewEvents(StartCause::WaitCancelled {
+                start: Instant::now(),
+                requested_resume: None,
+              }),
+              &window_target,
+              &mut control_flow,
+            );
+
+            for event in e.drain(..) {
+              match event {
+                Event::LoopDestroyed => control_flow = ControlFlow::Exit,
+                _ => callback(event, &window_target, &mut control_flow),
+              }
+            }
+
+            if control_flow != ControlFlow::Exit {
+              callback(Event::MainEventsCleared, &window_target, &mut control_flow);
+            }
+          }
+        }
+        ControlFlow::WaitUntil(requested_resume) => {
+          let mut e = events.lock().unwrap();
+          let start = Instant::now();
+          if start >= requested_resume {
+            callback(
+              Event::NewEvents(StartCause::ResumeTimeReached {
+                start,
+                requested_resume,
+              }),
+              &window_target,
+              &mut control_flow,
+            );
+
+            for event in e.drain(..) {
+              match event {
+                Event::LoopDestroyed => control_flow = ControlFlow::Exit,
+                _ => callback(event, &window_target, &mut control_flow),
+              }
+            }
+
+            if control_flow != ControlFlow::Exit {
+              callback(Event::MainEventsCleared, &window_target, &mut control_flow);
+            }
+          } else if !e.is_empty() {
+            callback(
+              Event::NewEvents(StartCause::WaitCancelled {
+                start,
+                requested_resume: Some(requested_resume),
+              }),
+              &window_target,
+              &mut control_flow,
+            );
+
+            for event in e.drain(..) {
+              match event {
+                Event::LoopDestroyed => control_flow = ControlFlow::Exit,
+                _ => callback(event, &window_target, &mut control_flow),
+              }
+            }
+
+            if control_flow != ControlFlow::Exit {
+              callback(Event::MainEventsCleared, &window_target, &mut control_flow);
+            }
+          }
+        }
+        ControlFlow::Poll => {
+          let mut e = events.lock().unwrap();
+          callback(
+            Event::NewEvents(StartCause::Poll),
+            &window_target,
+            &mut control_flow,
+          );
+          for event in e.drain(..) {
+            match event {
+              Event::LoopDestroyed => control_flow = ControlFlow::Exit,
+              _ => callback(event, &window_target, &mut control_flow),
+            }
+          }
+          callback(Event::MainEventsCleared, &window_target, &mut control_flow);
+        }
+      }
+
       gtk::main_iteration();
     }
     context.pop_thread_default();
