@@ -8,15 +8,16 @@ use std::{
   sync::atomic::{AtomicBool, AtomicI32, Ordering},
 };
 
-use gdk::{Cursor, EventMask, WindowEdge, WindowExt, WindowState};
+use gdk::{Cursor, EventKey, EventMask, WindowEdge, WindowExt, WindowState};
 use gdk_pixbuf::{Colorspace, Pixbuf};
 use gtk::{prelude::*, AccelGroup, ApplicationWindow, Orientation};
 
 use crate::{
   dpi::{PhysicalPosition, PhysicalSize, Position, Size},
   error::{ExternalError, NotSupportedError, OsError as RootOsError},
+  event::ElementState,
   icon::{BadIcon, Icon},
-  menu::{Menu, MenuItem},
+  menu::{MenuId, MenuItem},
   monitor::MonitorHandle as RootMonitorHandle,
   window::{CursorIcon, Fullscreen, UserAttentionType, WindowAttributes},
 };
@@ -84,12 +85,13 @@ pub struct Window {
   pub(crate) window_requests_tx: glib::Sender<(WindowId, WindowRequest)>,
   /// Gtk Acceleration Group
   pub(crate) accel_group: AccelGroup,
-  /// Gtk Menu
-  pub(crate) menu: gtk::Box,
+  // Gtk MenuBar allocation -- always available
+  menu_bar: gtk::MenuBar,
   scale_factor: Rc<AtomicI32>,
   position: Rc<(AtomicI32, AtomicI32)>,
   size: Rc<(AtomicI32, AtomicI32)>,
   maximized: Rc<AtomicBool>,
+  minimized: Rc<AtomicBool>,
   fullscreen: RefCell<Option<Fullscreen>>,
 }
 
@@ -184,12 +186,15 @@ impl Window {
       window.set_app_paintable(true);
     }
 
-    // Set Menu Bar
-    let menu = gtk::Box::new(Orientation::Vertical, 0);
-    if let Some(menus) = attributes.window_menu {
-      window.add(&menu);
-      let menubar = menu::initialize(window_id, menus, &window_requests_tx, &accel_group);
-      menu.pack_start(&menubar, false, false, 0);
+    // We always create a box and allocate menubar, so if they set_menu after creation
+    // we can inject the menubar without re-redendering the whole window
+    let window_box = gtk::Box::new(Orientation::Vertical, 0);
+    window.add(&window_box);
+
+    let mut menu_bar = gtk::MenuBar::new();
+    window_box.pack_start(&menu_bar, false, false, 0);
+    if let Some(window_menu) = attributes.window_menu {
+      window_menu.generate_menu(&mut menu_bar, &window_requests_tx, &accel_group, window_id);
     }
 
     // Rest attributes
@@ -245,6 +250,8 @@ impl Window {
       window.hide();
     }
 
+    window.set_skip_taskbar_hint(attributes.skip_taskbar);
+
     let w_pos = window.get_position();
     let position: Rc<(AtomicI32, AtomicI32)> = Rc::new((w_pos.0.into(), w_pos.1.into()));
     let position_clone = position.clone();
@@ -268,10 +275,13 @@ impl Window {
     let w_max = window.get_property_is_maximized();
     let maximized: Rc<AtomicBool> = Rc::new(w_max.into());
     let max_clone = maximized.clone();
+    let minimized = Rc::new(AtomicBool::new(false));
+    let min_clone = minimized.clone();
 
     window.connect_window_state_event(move |_window, event| {
       let state = event.get_new_window_state();
       max_clone.store(state.contains(WindowState::MAXIMIZED), Ordering::Release);
+      min_clone.store(state.contains(WindowState::ICONIFIED), Ordering::Release);
       Inhibit(false)
     });
 
@@ -281,21 +291,49 @@ impl Window {
       scale_factor_clone.store(window.get_scale_factor(), Ordering::Release);
     });
 
-    if let Err(e) = window_requests_tx.send((window_id, WindowRequest::WireUpEvents)) {
+    if let Err(e) = window_requests_tx
+      .clone()
+      .send((window_id, WindowRequest::WireUpEvents))
+    {
       log::warn!("Fail to send wire up events request: {}", e);
     }
 
+    let window_requests_tx_clone = window_requests_tx.clone();
+    window.connect_key_press_event(move |_window, event_key| {
+      if let Err(e) = window_requests_tx_clone.send((
+        window_id,
+        WindowRequest::KeyboardInput((event_key.to_owned(), ElementState::Pressed)),
+      )) {
+        log::warn!("Fail to send user attention request: {}", e);
+      }
+
+      Inhibit(false)
+    });
+
+    let window_requests_tx_clone = window_requests_tx.clone();
+    window.connect_key_release_event(move |_window, event_key| {
+      if let Err(e) = window_requests_tx_clone.send((
+        window_id,
+        WindowRequest::KeyboardInput((event_key.to_owned(), ElementState::Released)),
+      )) {
+        log::warn!("Fail to send user attention request: {}", e);
+      }
+
+      Inhibit(false)
+    });
+
     window.queue_draw();
     Ok(Self {
       window_id,
       window,
       window_requests_tx,
       accel_group,
-      menu,
+      menu_bar,
       scale_factor,
       position,
       size,
       maximized,
+      minimized,
       fullscreen: RefCell::new(attributes.fullscreen),
     })
   }
@@ -410,10 +448,10 @@ impl Window {
     }
   }
 
-  pub fn set_menu(&self, menu: Option<Vec<Menu>>) {
+  pub fn set_menu(&self, menu: Option<menu::Menu>) {
     if let Err(e) = self.window_requests_tx.send((
       self.window_id,
-      WindowRequest::SetMenu((menu, self.accel_group.clone(), self.menu.clone())),
+      WindowRequest::SetMenu((menu, self.accel_group.clone(), self.menu_bar.clone())),
     )) {
       log::warn!("Fail to send menu request: {}", e);
     }
@@ -429,11 +467,13 @@ impl Window {
   }
 
   pub fn set_focus(&self) {
-    if let Err(e) = self
-      .window_requests_tx
-      .send((self.window_id, WindowRequest::Focus))
-    {
-      log::warn!("Fail to send visible request: {}", e);
+    if !self.minimized.load(Ordering::Acquire) && self.window.get_visible() {
+      if let Err(e) = self
+        .window_requests_tx
+        .send((self.window_id, WindowRequest::Focus))
+      {
+        log::warn!("Fail to send visible request: {}", e);
+      }
     }
   }
 
@@ -476,6 +516,11 @@ impl Window {
     self.window.get_decorated()
   }
 
+  #[inline]
+  pub fn is_visible(&self) -> bool {
+    self.window.is_visible()
+  }
+
   pub fn drag_window(&self) -> Result<(), ExternalError> {
     if let Err(e) = self
       .window_requests_tx
@@ -540,6 +585,14 @@ impl Window {
     }
   }
 
+  pub fn hide_menu(&self) {
+    self.menu_bar.hide();
+  }
+
+  pub fn show_menu(&self) {
+    self.menu_bar.show_all();
+  }
+
   pub fn set_cursor_icon(&self, cursor: CursorIcon) {
     if let Err(e) = self
       .window_requests_tx
@@ -588,10 +641,10 @@ impl Window {
     todo!()
   }
 
-  pub fn skip_taskbar(&self) {
+  pub fn set_skip_taskbar(&self, skip: bool) {
     if let Err(e) = self
       .window_requests_tx
-      .send((self.window_id, WindowRequest::SkipTaskbar))
+      .send((self.window_id, WindowRequest::SetSkipTaskbar(skip)))
     {
       log::warn!("Fail to send skip taskbar request: {}", e);
     }
@@ -620,12 +673,14 @@ pub enum WindowRequest {
   AlwaysOnTop(bool),
   WindowIcon(Option<Icon>),
   UserAttention(Option<UserAttentionType>),
-  SkipTaskbar,
+  SetSkipTaskbar(bool),
   CursorIcon(Option<CursorIcon>),
   WireUpEvents,
   Redraw,
-  Menu(MenuItem),
-  SetMenu((Option<Vec<Menu>>, AccelGroup, gtk::Box)),
+  Menu((Option<MenuItem>, Option<MenuId>)),
+  SetMenu((Option<menu::Menu>, AccelGroup, gtk::MenuBar)),
+  KeyboardInput((EventKey, ElementState)),
+  GlobalHotKey(u16),
 }
 
 pub fn hit_test(window: &gdk::Window, cx: f64, cy: f64) -> WindowEdge {
