@@ -22,6 +22,7 @@ use std::{
 use winapi::shared::basetsd::{DWORD_PTR, UINT_PTR};
 
 use winapi::{
+  ctypes::c_int,
   shared::{
     minwindef::{BOOL, DWORD, HIWORD, INT, LOWORD, LPARAM, LRESULT, UINT, WORD, WPARAM},
     windef::{HWND, POINT, RECT},
@@ -30,20 +31,26 @@ use winapi::{
   um::{
     commctrl, libloaderapi, ole2, processthreadsapi, winbase,
     winnt::{HANDLE, LONG, LPCSTR, SHORT},
-    winuser,
+    winuser::{self, RAWINPUT},
   },
 };
 
 use crate::{
+  accelerator::AcceleratorId,
   dpi::{PhysicalPosition, PhysicalSize},
-  event::{DeviceEvent, Event, Force, KeyboardInput, Touch, TouchPhase, WindowEvent},
+  event::{DeviceEvent, Event, Force, RawKeyEvent, Touch, TouchPhase, WindowEvent},
   event_loop::{ControlFlow, EventLoopClosed, EventLoopWindowTarget as RootELW},
+  keyboard::{KeyCode, ModifiersState},
+  menu::{MenuId, MenuType},
   monitor::MonitorHandle as RootMonitorHandle,
   platform_impl::platform::{
+    accelerator,
     dark_mode::try_theme,
     dpi::{become_dpi_aware, dpi_to_scale_factor, enable_non_client_dpi_scaling},
     drop_handler::FileDropHandler,
-    event::{self, handle_extended_keys, process_key_params, vkey_to_tao_vkey},
+    keyboard::is_msg_keyboard_related,
+    keyboard_layout::LAYOUT_CACHE,
+    minimal_ime::is_msg_ime_related,
     monitor::{self, MonitorHandle},
     raw_input, util,
     window_state::{CursorFlags, WindowFlags, WindowState},
@@ -111,6 +118,13 @@ impl<T> ThreadMsgTargetSubclassInput<T> {
   }
 }
 
+/// The result of a subclass procedure (the message handling callback)
+pub(crate) enum ProcResult {
+  DefSubclassProc, // <- this should be the default value
+  DefWindowProc,
+  Value(isize),
+}
+
 pub struct EventLoop<T: 'static> {
   thread_msg_sender: Sender<T>,
   window_target: RootELW<T>,
@@ -220,8 +234,23 @@ impl<T: 'static> EventLoop<T> {
         if 0 == winuser::GetMessageW(&mut msg, ptr::null_mut(), 0, 0) {
           break 'main;
         }
-        winuser::TranslateMessage(&mut msg);
-        winuser::DispatchMessageW(&mut msg);
+
+        // global accelerator
+        if msg.message == winuser::WM_HOTKEY {
+          let event_loop_runner = self.window_target.p.runner_shared.clone();
+          event_loop_runner
+            .send_event(Event::GlobalShortcutEvent(AcceleratorId(msg.wParam as u16)));
+        }
+
+        // window accelerator
+        let accels = accelerator::find_accels(winuser::GetAncestor(msg.hwnd, winuser::GA_ROOT));
+        let translated = accels.map_or(false, |it| {
+          winuser::TranslateAcceleratorW(msg.hwnd, it.handle(), &mut msg) != 0
+        });
+        if !translated {
+          winuser::TranslateMessage(&mut msg);
+          winuser::DispatchMessageW(&mut msg);
+        }
 
         if let Err(payload) = runner.take_panic_error() {
           runner.reset_runner();
@@ -294,7 +323,7 @@ fn get_wait_thread_id() -> DWORD {
     );
     assert_eq!(
       msg.message, *SEND_WAIT_THREAD_ID_MSG_ID,
-      "this shouldn't be possible. please open an issue with Tao. error code: {}",
+      "this shouldn't be possible. please open an issue with Tauri. error code: {}",
       result
     );
     msg.lParam as DWORD
@@ -751,10 +780,15 @@ unsafe fn process_control_flow<T: 'static>(runner: &EventLoopRunner<T>) {
 }
 
 /// Emit a `ModifiersChanged` event whenever modifiers have changed.
-fn update_modifiers<T>(window: HWND, subclass_input: &SubclassInput<T>) {
+/// Returns the current modifier state
+fn update_modifiers<T>(window: HWND, subclass_input: &SubclassInput<T>) -> ModifiersState {
   use crate::event::WindowEvent::ModifiersChanged;
 
-  let modifiers = event::get_key_mods();
+  let modifiers = {
+    let mut layouts = LAYOUT_CACHE.lock().unwrap();
+    layouts.get_agnostic_mods()
+  };
+
   let mut window_state = subclass_input.window_state.lock();
   if window_state.modifiers_state != modifiers {
     window_state.modifiers_state = modifiers;
@@ -769,6 +803,7 @@ fn update_modifiers<T>(window: HWND, subclass_input: &SubclassInput<T>) {
       });
     }
   }
+  modifiers
 }
 
 /// Any window whose callback is configured to this function will have its events propagated
@@ -825,6 +860,75 @@ unsafe fn public_window_callback_inner<T: 'static>(
     winuser::RDW_INTERNALPAINT,
   );
 
+  let mut result = ProcResult::DefSubclassProc;
+
+  // Send new modifiers before sending key events.
+  let mods_changed_callback = || match msg {
+    winuser::WM_KEYDOWN | winuser::WM_SYSKEYDOWN | winuser::WM_KEYUP | winuser::WM_SYSKEYUP => {
+      update_modifiers(window, subclass_input);
+      result = ProcResult::Value(0);
+    }
+    _ => (),
+  };
+  subclass_input
+    .event_loop_runner
+    .catch_unwind(mods_changed_callback)
+    .unwrap_or_else(|| result = ProcResult::Value(-1));
+
+  let keyboard_callback = || {
+    use crate::event::WindowEvent::KeyboardInput;
+    let is_keyboard_related = is_msg_keyboard_related(msg);
+    if !is_keyboard_related {
+      // We return early to avoid a deadlock from locking the window state
+      // when not appropriate.
+      return;
+    }
+    let events = {
+      let mut window_state = subclass_input.window_state.lock();
+      window_state
+        .key_event_builder
+        .process_message(window, msg, wparam, lparam, &mut result)
+    };
+    for event in events {
+      subclass_input.send_event(Event::WindowEvent {
+        window_id: RootWindowId(WindowId(window)),
+        event: KeyboardInput {
+          device_id: DEVICE_ID,
+          event: event.event,
+          is_synthetic: event.is_synthetic,
+        },
+      });
+    }
+  };
+  subclass_input
+    .event_loop_runner
+    .catch_unwind(keyboard_callback)
+    .unwrap_or_else(|| result = ProcResult::Value(-1));
+
+  let ime_callback = || {
+    use crate::event::WindowEvent::ReceivedImeText;
+    let is_ime_related = is_msg_ime_related(msg);
+    if !is_ime_related {
+      return;
+    }
+    let text = {
+      let mut window_state = subclass_input.window_state.lock();
+      window_state
+        .ime_handler
+        .process_message(window, msg, wparam, lparam, &mut result)
+    };
+    if let Some(str) = text {
+      subclass_input.send_event(Event::WindowEvent {
+        window_id: RootWindowId(WindowId(window)),
+        event: ReceivedImeText(str),
+      });
+    }
+  };
+  subclass_input
+    .event_loop_runner
+    .catch_unwind(ime_callback)
+    .unwrap_or_else(|| result = ProcResult::Value(-1));
+
   // I decided to bind the closure to `callback` and pass it to catch_unwind rather than passing
   // the closure to catch_unwind directly so that the match body indendation wouldn't change and
   // the git blame and history would be preserved.
@@ -834,7 +938,7 @@ unsafe fn public_window_callback_inner<T: 'static>(
         .window_state
         .lock()
         .set_window_flags_in_place(|f| f.insert(WindowFlags::MARKER_IN_SIZE_MOVE));
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_EXITSIZEMOVE => {
@@ -842,18 +946,16 @@ unsafe fn public_window_callback_inner<T: 'static>(
         .window_state
         .lock()
         .set_window_flags_in_place(|f| f.remove(WindowFlags::MARKER_IN_SIZE_MOVE));
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_NCCREATE => {
       enable_non_client_dpi_scaling(window);
-      commctrl::DefSubclassProc(window, msg, wparam, lparam)
     }
     winuser::WM_NCLBUTTONDOWN => {
       if wparam == winuser::HTCAPTION as _ {
         winuser::PostMessageW(window, winuser::WM_MOUSEMOVE, 0, lparam);
       }
-      commctrl::DefSubclassProc(window, msg, wparam, lparam)
     }
 
     winuser::WM_CLOSE => {
@@ -862,7 +964,7 @@ unsafe fn public_window_callback_inner<T: 'static>(
         window_id: RootWindowId(WindowId(window)),
         event: CloseRequested,
       });
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_DESTROY => {
@@ -873,13 +975,13 @@ unsafe fn public_window_callback_inner<T: 'static>(
         event: Destroyed,
       });
       subclass_input.event_loop_runner.remove_window(window);
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_NCDESTROY => {
       remove_window_subclass::<T>(window);
       subclass_input.subclass_removed.set(true);
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_PAINT => {
@@ -900,8 +1002,6 @@ unsafe fn public_window_callback_inner<T: 'static>(
           process_control_flow(&subclass_input.event_loop_runner);
         }
       }
-
-      commctrl::DefSubclassProc(window, msg, wparam, lparam)
     }
 
     winuser::WM_WINDOWPOSCHANGING => {
@@ -948,7 +1048,7 @@ unsafe fn public_window_callback_inner<T: 'static>(
         }
       }
 
-      0
+      result = ProcResult::Value(0);
     }
 
     // WM_MOVE supplies client area positions, so we send Moved here instead.
@@ -965,7 +1065,7 @@ unsafe fn public_window_callback_inner<T: 'static>(
       }
 
       // This is necessary for us to still get sent WM_SIZE.
-      commctrl::DefSubclassProc(window, msg, wparam, lparam)
+      result = ProcResult::DefSubclassProc;
     }
 
     winuser::WM_SIZE => {
@@ -992,40 +1092,17 @@ unsafe fn public_window_callback_inner<T: 'static>(
       }
 
       subclass_input.send_event(event);
-      0
+      result = ProcResult::Value(0);
     }
 
-    winuser::WM_CHAR | winuser::WM_SYSCHAR => {
-      use crate::event::WindowEvent::ReceivedCharacter;
-      use std::char;
-      let is_high_surrogate = 0xD800 <= wparam && wparam <= 0xDBFF;
-      let is_low_surrogate = 0xDC00 <= wparam && wparam <= 0xDFFF;
-
-      if is_high_surrogate {
-        subclass_input.window_state.lock().high_surrogate = Some(wparam as u16);
-      } else if is_low_surrogate {
-        let high_surrogate = subclass_input.window_state.lock().high_surrogate.take();
-
-        if let Some(high_surrogate) = high_surrogate {
-          let pair = [high_surrogate, wparam as u16];
-          if let Some(Ok(chr)) = char::decode_utf16(pair.iter().copied()).next() {
-            subclass_input.send_event(Event::WindowEvent {
-              window_id: RootWindowId(WindowId(window)),
-              event: ReceivedCharacter(chr),
-            });
-          }
-        }
-      } else {
-        subclass_input.window_state.lock().high_surrogate = None;
-
-        if let Some(chr) = char::from_u32(wparam as u32) {
-          subclass_input.send_event(Event::WindowEvent {
-            window_id: RootWindowId(WindowId(window)),
-            event: ReceivedCharacter(chr),
-          });
-        }
-      }
-      0
+    // this is catched by the accelerator translator
+    winuser::WM_COMMAND => {
+      let menu_id = LOWORD(wparam as u32) as u16;
+      subclass_input.send_event(Event::MenuEvent {
+        menu_id: MenuId(menu_id as u16),
+        // todo fix menutype
+        origin: MenuType::MenuBar,
+      });
     }
 
     // this is necessary for us to maintain minimize/restore state
@@ -1043,11 +1120,12 @@ unsafe fn public_window_callback_inner<T: 'static>(
       if wparam == winuser::SC_SCREENSAVE {
         let window_state = subclass_input.window_state.lock();
         if window_state.fullscreen.is_some() {
-          return 0;
+          result = ProcResult::Value(0);
+          return;
         }
       }
 
-      winuser::DefWindowProcW(window, msg, wparam, lparam)
+      result = ProcResult::DefWindowProc;
     }
 
     winuser::WM_MOUSEMOVE => {
@@ -1092,19 +1170,18 @@ unsafe fn public_window_callback_inner<T: 'static>(
         w.mouse.last_position = Some(position);
       }
       if cursor_moved {
-        update_modifiers(window, subclass_input);
-
+        let modifiers = update_modifiers(window, subclass_input);
         subclass_input.send_event(Event::WindowEvent {
           window_id: RootWindowId(WindowId(window)),
           event: CursorMoved {
             device_id: DEVICE_ID,
             position,
-            modifiers: event::get_key_mods(),
+            modifiers,
           },
         });
       }
 
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_MOUSELEAVE => {
@@ -1123,7 +1200,7 @@ unsafe fn public_window_callback_inner<T: 'static>(
         },
       });
 
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_MOUSEWHEEL => {
@@ -1133,7 +1210,7 @@ unsafe fn public_window_callback_inner<T: 'static>(
       let value = value as i32;
       let value = value as f32 / winuser::WHEEL_DELTA as f32;
 
-      update_modifiers(window, subclass_input);
+      let modifiers = update_modifiers(window, subclass_input);
 
       subclass_input.send_event(Event::WindowEvent {
         window_id: RootWindowId(WindowId(window)),
@@ -1141,11 +1218,11 @@ unsafe fn public_window_callback_inner<T: 'static>(
           device_id: DEVICE_ID,
           delta: LineDelta(0.0, value),
           phase: TouchPhase::Moved,
-          modifiers: event::get_key_mods(),
+          modifiers,
         },
       });
 
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_MOUSEHWHEEL => {
@@ -1155,7 +1232,7 @@ unsafe fn public_window_callback_inner<T: 'static>(
       let value = value as i32;
       let value = value as f32 / winuser::WHEEL_DELTA as f32;
 
-      update_modifiers(window, subclass_input);
+      let modifiers = update_modifiers(window, subclass_input);
 
       subclass_input.send_event(Event::WindowEvent {
         window_id: RootWindowId(WindowId(window)),
@@ -1163,77 +1240,25 @@ unsafe fn public_window_callback_inner<T: 'static>(
           device_id: DEVICE_ID,
           delta: LineDelta(value, 0.0),
           phase: TouchPhase::Moved,
-          modifiers: event::get_key_mods(),
+          modifiers,
         },
       });
 
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_KEYDOWN | winuser::WM_SYSKEYDOWN => {
-      use crate::event::{ElementState::Pressed, VirtualKeyCode};
       if msg == winuser::WM_SYSKEYDOWN && wparam as i32 == winuser::VK_F4 {
-        commctrl::DefSubclassProc(window, msg, wparam, lparam)
-      } else {
-        if let Some((scancode, vkey)) = process_key_params(wparam, lparam) {
-          update_modifiers(window, subclass_input);
-
-          #[allow(deprecated)]
-          subclass_input.send_event(Event::WindowEvent {
-            window_id: RootWindowId(WindowId(window)),
-            event: WindowEvent::KeyboardInput {
-              device_id: DEVICE_ID,
-              input: KeyboardInput {
-                state: Pressed,
-                scancode,
-                virtual_keycode: vkey,
-                modifiers: event::get_key_mods(),
-              },
-              is_synthetic: false,
-            },
-          });
-          // Windows doesn't emit a delete character by default, but in order to make it
-          // consistent with the other platforms we'll emit a delete character here.
-          if vkey == Some(VirtualKeyCode::Delete) {
-            subclass_input.send_event(Event::WindowEvent {
-              window_id: RootWindowId(WindowId(window)),
-              event: WindowEvent::ReceivedCharacter('\u{7F}'),
-            });
-          }
-        }
-        0
+        result = ProcResult::DefSubclassProc;
       }
     }
 
-    winuser::WM_KEYUP | winuser::WM_SYSKEYUP => {
-      use crate::event::ElementState::Released;
-      if let Some((scancode, vkey)) = process_key_params(wparam, lparam) {
-        update_modifiers(window, subclass_input);
-
-        #[allow(deprecated)]
-        subclass_input.send_event(Event::WindowEvent {
-          window_id: RootWindowId(WindowId(window)),
-          event: WindowEvent::KeyboardInput {
-            device_id: DEVICE_ID,
-            input: KeyboardInput {
-              state: Released,
-              scancode,
-              virtual_keycode: vkey,
-              modifiers: event::get_key_mods(),
-            },
-            is_synthetic: false,
-          },
-        });
-      }
-      0
-    }
-
     winuser::WM_LBUTTONDOWN => {
       use crate::event::{ElementState::Pressed, MouseButton::Left, WindowEvent::MouseInput};
 
       capture_mouse(window, &mut *subclass_input.window_state.lock());
 
-      update_modifiers(window, subclass_input);
+      let modifiers = update_modifiers(window, subclass_input);
 
       subclass_input.send_event(Event::WindowEvent {
         window_id: RootWindowId(WindowId(window)),
@@ -1241,10 +1266,10 @@ unsafe fn public_window_callback_inner<T: 'static>(
           device_id: DEVICE_ID,
           state: Pressed,
           button: Left,
-          modifiers: event::get_key_mods(),
+          modifiers,
         },
       });
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_LBUTTONUP => {
@@ -1252,7 +1277,7 @@ unsafe fn public_window_callback_inner<T: 'static>(
 
       release_mouse(subclass_input.window_state.lock());
 
-      update_modifiers(window, subclass_input);
+      let modifiers = update_modifiers(window, subclass_input);
 
       subclass_input.send_event(Event::WindowEvent {
         window_id: RootWindowId(WindowId(window)),
@@ -1260,10 +1285,10 @@ unsafe fn public_window_callback_inner<T: 'static>(
           device_id: DEVICE_ID,
           state: Released,
           button: Left,
-          modifiers: event::get_key_mods(),
+          modifiers,
         },
       });
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_RBUTTONDOWN => {
@@ -1271,7 +1296,7 @@ unsafe fn public_window_callback_inner<T: 'static>(
 
       capture_mouse(window, &mut *subclass_input.window_state.lock());
 
-      update_modifiers(window, subclass_input);
+      let modifiers = update_modifiers(window, subclass_input);
 
       subclass_input.send_event(Event::WindowEvent {
         window_id: RootWindowId(WindowId(window)),
@@ -1279,10 +1304,10 @@ unsafe fn public_window_callback_inner<T: 'static>(
           device_id: DEVICE_ID,
           state: Pressed,
           button: Right,
-          modifiers: event::get_key_mods(),
+          modifiers,
         },
       });
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_RBUTTONUP => {
@@ -1290,7 +1315,7 @@ unsafe fn public_window_callback_inner<T: 'static>(
 
       release_mouse(subclass_input.window_state.lock());
 
-      update_modifiers(window, subclass_input);
+      let modifiers = update_modifiers(window, subclass_input);
 
       subclass_input.send_event(Event::WindowEvent {
         window_id: RootWindowId(WindowId(window)),
@@ -1298,10 +1323,10 @@ unsafe fn public_window_callback_inner<T: 'static>(
           device_id: DEVICE_ID,
           state: Released,
           button: Right,
-          modifiers: event::get_key_mods(),
+          modifiers,
         },
       });
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_MBUTTONDOWN => {
@@ -1309,7 +1334,7 @@ unsafe fn public_window_callback_inner<T: 'static>(
 
       capture_mouse(window, &mut *subclass_input.window_state.lock());
 
-      update_modifiers(window, subclass_input);
+      let modifiers = update_modifiers(window, subclass_input);
 
       subclass_input.send_event(Event::WindowEvent {
         window_id: RootWindowId(WindowId(window)),
@@ -1317,10 +1342,10 @@ unsafe fn public_window_callback_inner<T: 'static>(
           device_id: DEVICE_ID,
           state: Pressed,
           button: Middle,
-          modifiers: event::get_key_mods(),
+          modifiers,
         },
       });
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_MBUTTONUP => {
@@ -1328,7 +1353,7 @@ unsafe fn public_window_callback_inner<T: 'static>(
 
       release_mouse(subclass_input.window_state.lock());
 
-      update_modifiers(window, subclass_input);
+      let modifiers = update_modifiers(window, subclass_input);
 
       subclass_input.send_event(Event::WindowEvent {
         window_id: RootWindowId(WindowId(window)),
@@ -1336,10 +1361,10 @@ unsafe fn public_window_callback_inner<T: 'static>(
           device_id: DEVICE_ID,
           state: Released,
           button: Middle,
-          modifiers: event::get_key_mods(),
+          modifiers,
         },
       });
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_XBUTTONDOWN => {
@@ -1348,7 +1373,7 @@ unsafe fn public_window_callback_inner<T: 'static>(
 
       capture_mouse(window, &mut *subclass_input.window_state.lock());
 
-      update_modifiers(window, subclass_input);
+      let modifiers = update_modifiers(window, subclass_input);
 
       subclass_input.send_event(Event::WindowEvent {
         window_id: RootWindowId(WindowId(window)),
@@ -1356,10 +1381,10 @@ unsafe fn public_window_callback_inner<T: 'static>(
           device_id: DEVICE_ID,
           state: Pressed,
           button: Other(xbutton),
-          modifiers: event::get_key_mods(),
+          modifiers,
         },
       });
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_XBUTTONUP => {
@@ -1368,7 +1393,7 @@ unsafe fn public_window_callback_inner<T: 'static>(
 
       release_mouse(subclass_input.window_state.lock());
 
-      update_modifiers(window, subclass_input);
+      let modifiers = update_modifiers(window, subclass_input);
 
       subclass_input.send_event(Event::WindowEvent {
         window_id: RootWindowId(WindowId(window)),
@@ -1376,10 +1401,10 @@ unsafe fn public_window_callback_inner<T: 'static>(
           device_id: DEVICE_ID,
           state: Released,
           button: Other(xbutton),
-          modifiers: event::get_key_mods(),
+          modifiers,
         },
       });
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_CAPTURECHANGED => {
@@ -1390,7 +1415,7 @@ unsafe fn public_window_callback_inner<T: 'static>(
       if lparam != window as isize {
         subclass_input.window_state.lock().mouse.capture_count = 0;
       }
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_TOUCH => {
@@ -1439,7 +1464,7 @@ unsafe fn public_window_callback_inner<T: 'static>(
         }
       }
       winuser::CloseTouchInputHandle(htouch);
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_POINTERDOWN | winuser::WM_POINTERUPDATE | winuser::WM_POINTERUP => {
@@ -1462,7 +1487,8 @@ unsafe fn public_window_callback_inner<T: 'static>(
           std::ptr::null_mut(),
         ) == 0
         {
-          return 0;
+          result = ProcResult::Value(0);
+          return;
         }
 
         let pointer_info_count = (entries_count * pointers_count) as usize;
@@ -1475,7 +1501,8 @@ unsafe fn public_window_callback_inner<T: 'static>(
           pointer_infos.as_mut_ptr(),
         ) == 0
         {
-          return 0;
+          result = ProcResult::Value(0);
+          return;
         }
 
         // https://docs.microsoft.com/en-us/windows/desktop/api/winuser/nf-winuser-getpointerframeinfohistory
@@ -1568,66 +1595,23 @@ unsafe fn public_window_callback_inner<T: 'static>(
 
         SkipPointerFrameMessages(pointer_id);
       }
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_SETFOCUS => {
-      use crate::event::{ElementState::Released, WindowEvent::Focused};
-      for windows_keycode in event::get_pressed_keys() {
-        let scancode = winuser::MapVirtualKeyA(windows_keycode as _, winuser::MAPVK_VK_TO_VSC);
-        let virtual_keycode = event::vkey_to_tao_vkey(windows_keycode);
-
-        update_modifiers(window, subclass_input);
-
-        #[allow(deprecated)]
-        subclass_input.send_event(Event::WindowEvent {
-          window_id: RootWindowId(WindowId(window)),
-          event: WindowEvent::KeyboardInput {
-            device_id: DEVICE_ID,
-            input: KeyboardInput {
-              scancode,
-              virtual_keycode,
-              state: Released,
-              modifiers: event::get_key_mods(),
-            },
-            is_synthetic: true,
-          },
-        })
-      }
+      use crate::event::WindowEvent::Focused;
+      update_modifiers(window, subclass_input);
 
       subclass_input.send_event(Event::WindowEvent {
         window_id: RootWindowId(WindowId(window)),
         event: Focused(true),
       });
 
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_KILLFOCUS => {
-      use crate::event::{
-        ElementState::Released,
-        ModifiersState,
-        WindowEvent::{Focused, ModifiersChanged},
-      };
-      for windows_keycode in event::get_pressed_keys() {
-        let scancode = winuser::MapVirtualKeyA(windows_keycode as _, winuser::MAPVK_VK_TO_VSC);
-        let virtual_keycode = event::vkey_to_tao_vkey(windows_keycode);
-
-        #[allow(deprecated)]
-        subclass_input.send_event(Event::WindowEvent {
-          window_id: RootWindowId(WindowId(window)),
-          event: WindowEvent::KeyboardInput {
-            device_id: DEVICE_ID,
-            input: KeyboardInput {
-              scancode,
-              virtual_keycode,
-              state: Released,
-              modifiers: event::get_key_mods(),
-            },
-            is_synthetic: true,
-          },
-        })
-      }
+      use crate::event::WindowEvent::{Focused, ModifiersChanged};
 
       subclass_input.window_state.lock().modifiers_state = ModifiersState::empty();
       subclass_input.send_event(Event::WindowEvent {
@@ -1639,7 +1623,7 @@ unsafe fn public_window_callback_inner<T: 'static>(
         window_id: RootWindowId(WindowId(window)),
         event: Focused(false),
       });
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_SETCURSOR => {
@@ -1660,17 +1644,12 @@ unsafe fn public_window_callback_inner<T: 'static>(
         Some(cursor) => {
           let cursor = winuser::LoadCursorW(ptr::null_mut(), cursor.to_windows_cursor());
           winuser::SetCursor(cursor);
-          0
+          result = ProcResult::Value(0);
         }
-        None => winuser::DefWindowProcW(window, msg, wparam, lparam),
+        None => result = ProcResult::DefWindowProc,
       }
     }
 
-    winuser::WM_DROPFILES => {
-      // See `FileDropHandler` for implementation.
-      0
-    }
-
     winuser::WM_GETMINMAXINFO => {
       let mmi = lparam as *mut winuser::MINMAXINFO;
 
@@ -1695,7 +1674,7 @@ unsafe fn public_window_callback_inner<T: 'static>(
         }
       }
 
-      0
+      result = ProcResult::Value(0);
     }
 
     // Only sent on Windows 8.1 or newer. On Windows 7 and older user has to log out to change
@@ -1717,7 +1696,8 @@ unsafe fn public_window_callback_inner<T: 'static>(
         window_state.scale_factor = new_scale_factor;
 
         if new_scale_factor == old_scale_factor {
-          return 0;
+          result = ProcResult::Value(0);
+          return;
         }
 
         window_state.fullscreen.is_none()
@@ -1907,7 +1887,7 @@ unsafe fn public_window_callback_inner<T: 'static>(
         winuser::SWP_NOZORDER | winuser::SWP_NOACTIVATE,
       );
 
-      0
+      result = ProcResult::Value(0);
     }
 
     winuser::WM_SETTINGCHANGE => {
@@ -1928,22 +1908,18 @@ unsafe fn public_window_callback_inner<T: 'static>(
           });
         }
       }
-
-      commctrl::DefSubclassProc(window, msg, wparam, lparam)
     }
 
     _ => {
       if msg == *DESTROY_MSG_ID {
         winuser::DestroyWindow(window);
-        0
+        result = ProcResult::Value(0);
       } else if msg == *SET_RETAIN_STATE_ON_SIZE_MSG_ID {
         let mut window_state = subclass_input.window_state.lock();
         window_state.set_window_flags_in_place(|f| {
           f.set(WindowFlags::MARKER_RETAIN_STATE_ON_SIZE, wparam != 0)
         });
-        0
-      } else {
-        commctrl::DefSubclassProc(window, msg, wparam, lparam)
+        result = ProcResult::Value(0);
       }
     }
   };
@@ -1951,7 +1927,13 @@ unsafe fn public_window_callback_inner<T: 'static>(
   subclass_input
     .event_loop_runner
     .catch_unwind(callback)
-    .unwrap_or(-1)
+    .unwrap_or_else(|| result = ProcResult::Value(-1));
+
+  match result {
+    ProcResult::DefSubclassProc => commctrl::DefSubclassProc(window, msg, wparam, lparam),
+    ProcResult::DefWindowProc => winuser::DefWindowProcW(window, msg, wparam, lparam),
+    ProcResult::Value(val) => val,
+  }
 }
 
 unsafe extern "system" fn thread_event_target_callback<T: 'static>(
@@ -2033,101 +2015,8 @@ unsafe extern "system" fn thread_event_target_callback<T: 'static>(
     }
 
     winuser::WM_INPUT => {
-      use crate::event::{
-        DeviceEvent::{Button, Key, Motion, MouseMotion, MouseWheel},
-        ElementState::{Pressed, Released},
-        MouseScrollDelta::LineDelta,
-      };
-
       if let Some(data) = raw_input::get_raw_input_data(lparam as _) {
-        let device_id = wrap_device_id(data.header.hDevice as _);
-
-        if data.header.dwType == winuser::RIM_TYPEMOUSE {
-          let mouse = data.data.mouse();
-
-          if util::has_flag(mouse.usFlags, winuser::MOUSE_MOVE_RELATIVE) {
-            let x = mouse.lLastX as f64;
-            let y = mouse.lLastY as f64;
-
-            if x != 0.0 {
-              subclass_input.send_event(Event::DeviceEvent {
-                device_id,
-                event: Motion { axis: 0, value: x },
-              });
-            }
-
-            if y != 0.0 {
-              subclass_input.send_event(Event::DeviceEvent {
-                device_id,
-                event: Motion { axis: 1, value: y },
-              });
-            }
-
-            if x != 0.0 || y != 0.0 {
-              subclass_input.send_event(Event::DeviceEvent {
-                device_id,
-                event: MouseMotion { delta: (x, y) },
-              });
-            }
-          }
-
-          if util::has_flag(mouse.usButtonFlags, winuser::RI_MOUSE_WHEEL) {
-            let delta = mouse.usButtonData as SHORT as f32 / winuser::WHEEL_DELTA as f32;
-            subclass_input.send_event(Event::DeviceEvent {
-              device_id,
-              event: MouseWheel {
-                delta: LineDelta(0.0, delta),
-              },
-            });
-          }
-
-          let button_state = raw_input::get_raw_mouse_button_state(mouse.usButtonFlags);
-          // Left, middle, and right, respectively.
-          for (index, state) in button_state.iter().enumerate() {
-            if let Some(state) = *state {
-              // This gives us consistency with Linux, since there doesn't
-              // seem to be anything else reasonable to do for a mouse
-              // button ID.
-              let button = (index + 1) as _;
-              subclass_input.send_event(Event::DeviceEvent {
-                device_id,
-                event: Button { button, state },
-              });
-            }
-          }
-        } else if data.header.dwType == winuser::RIM_TYPEKEYBOARD {
-          let keyboard = data.data.keyboard();
-
-          let pressed =
-            keyboard.Message == winuser::WM_KEYDOWN || keyboard.Message == winuser::WM_SYSKEYDOWN;
-          let released =
-            keyboard.Message == winuser::WM_KEYUP || keyboard.Message == winuser::WM_SYSKEYUP;
-
-          if pressed || released {
-            let state = if pressed { Pressed } else { Released };
-
-            let scancode = keyboard.MakeCode as _;
-            let extended = util::has_flag(keyboard.Flags, winuser::RI_KEY_E0 as _)
-              | util::has_flag(keyboard.Flags, winuser::RI_KEY_E1 as _);
-
-            if let Some((vkey, scancode)) =
-              handle_extended_keys(keyboard.VKey as _, scancode, extended)
-            {
-              let virtual_keycode = vkey_to_tao_vkey(vkey);
-
-              #[allow(deprecated)]
-              subclass_input.send_event(Event::DeviceEvent {
-                device_id,
-                event: Key(KeyboardInput {
-                  scancode,
-                  state,
-                  virtual_keycode,
-                  modifiers: event::get_key_mods(),
-                }),
-              });
-            }
-          }
-        }
+        handle_raw_input(&subclass_input, data);
       }
 
       commctrl::DefSubclassProc(window, msg, wparam, lparam)
@@ -2197,3 +2086,183 @@ unsafe extern "system" fn thread_event_target_callback<T: 'static>(
   }
   result
 }
+
+unsafe fn handle_raw_input<T: 'static>(
+  subclass_input: &Box<ThreadMsgTargetSubclassInput<T>>,
+  data: RAWINPUT,
+) {
+  use crate::event::{
+    DeviceEvent::{Button, Key, Motion, MouseMotion, MouseWheel},
+    ElementState::{Pressed, Released},
+    MouseScrollDelta::LineDelta,
+  };
+
+  let device_id = wrap_device_id(data.header.hDevice as _);
+
+  if data.header.dwType == winuser::RIM_TYPEMOUSE {
+    let mouse = data.data.mouse();
+
+    if util::has_flag(mouse.usFlags, winuser::MOUSE_MOVE_RELATIVE) {
+      let x = mouse.lLastX as f64;
+      let y = mouse.lLastY as f64;
+
+      if x != 0.0 {
+        subclass_input.send_event(Event::DeviceEvent {
+          device_id,
+          event: Motion { axis: 0, value: x },
+        });
+      }
+
+      if y != 0.0 {
+        subclass_input.send_event(Event::DeviceEvent {
+          device_id,
+          event: Motion { axis: 1, value: y },
+        });
+      }
+
+      if x != 0.0 || y != 0.0 {
+        subclass_input.send_event(Event::DeviceEvent {
+          device_id,
+          event: MouseMotion { delta: (x, y) },
+        });
+      }
+    }
+
+    if util::has_flag(mouse.usButtonFlags, winuser::RI_MOUSE_WHEEL) {
+      // We must cast to SHORT first, becaues `usButtonData` must be interpreted as signed.
+      let delta = mouse.usButtonData as SHORT as f32 / winuser::WHEEL_DELTA as f32;
+      subclass_input.send_event(Event::DeviceEvent {
+        device_id,
+        event: MouseWheel {
+          delta: LineDelta(0.0, delta),
+        },
+      });
+    }
+
+    let button_state = raw_input::get_raw_mouse_button_state(mouse.usButtonFlags);
+    // Left, middle, and right, respectively.
+    for (index, state) in button_state.iter().enumerate() {
+      if let Some(state) = *state {
+        // This gives us consistency with X11, since there doesn't
+        // seem to be anything else reasonable to do for a mouse
+        // button ID.
+        let button = (index + 1) as _;
+        subclass_input.send_event(Event::DeviceEvent {
+          device_id,
+          event: Button { button, state },
+        });
+      }
+    }
+  } else if data.header.dwType == winuser::RIM_TYPEKEYBOARD {
+    let keyboard = data.data.keyboard();
+
+    let pressed =
+      keyboard.Message == winuser::WM_KEYDOWN || keyboard.Message == winuser::WM_SYSKEYDOWN;
+    let released =
+      keyboard.Message == winuser::WM_KEYUP || keyboard.Message == winuser::WM_SYSKEYUP;
+
+    if !pressed && !released {
+      return;
+    }
+
+    let state = if pressed { Pressed } else { Released };
+    let extension = {
+      if util::has_flag(keyboard.Flags, winuser::RI_KEY_E0 as _) {
+        0xE000
+      } else if util::has_flag(keyboard.Flags, winuser::RI_KEY_E1 as _) {
+        0xE100
+      } else {
+        0x0000
+      }
+    };
+    let scancode;
+    if keyboard.MakeCode == 0 {
+      // In some cases (often with media keys) the device reports a scancode of 0 but a
+      // valid virtual key. In these cases we obtain the scancode from the virtual key.
+      scancode = winuser::MapVirtualKeyW(keyboard.VKey as u32, winuser::MAPVK_VK_TO_VSC_EX) as u16;
+    } else {
+      scancode = keyboard.MakeCode | extension;
+    }
+    if scancode == 0xE11D || scancode == 0xE02A {
+      // At the hardware (or driver?) level, pressing the Pause key is equivalent to pressing
+      // Ctrl+NumLock.
+      // This equvalence means that if the user presses Pause, the keyboard will emit two
+      // subsequent keypresses:
+      // 1, 0xE11D - Which is a left Ctrl (0x1D) with an extension flag (0xE100)
+      // 2, 0x0045 - Which on its own can be interpreted as Pause
+      //
+      // There's another combination which isn't quite an equivalence:
+      // PrtSc used to be Shift+Asterisk. This means that on some keyboards, presssing
+      // PrtSc (print screen) produces the following sequence:
+      // 1, 0xE02A - Which is a left shift (0x2A) with an extension flag (0xE000)
+      // 2, 0xE037 - Which is a numpad multiply (0x37) with an exteion flag (0xE000). This on
+      //             its own it can be interpreted as PrtSc
+      //
+      // For this reason, if we encounter the first keypress, we simply ignore it, trusting
+      // that there's going to be another event coming, from which we can extract the
+      // appropriate key.
+      // For more on this, read the article by Raymond Chen, titled:
+      // "Why does Ctrl+ScrollLock cancel dialogs?"
+      // https://devblogs.microsoft.com/oldnewthing/20080211-00/?p=23503
+      return;
+    }
+    let code;
+    if keyboard.VKey as c_int == winuser::VK_NUMLOCK {
+      // Historically, the NumLock and the Pause key were one and the same physical key.
+      // The user could trigger Pause by pressing Ctrl+NumLock.
+      // Now these are often physically separate and the two keys can be differentiated by
+      // checking the extension flag of the scancode. NumLock is 0xE045, Pause is 0x0045.
+      //
+      // However in this event, both keys are reported as 0x0045 even on modern hardware.
+      // Therefore we use the virtual key instead to determine whether it's a NumLock and
+      // set the KeyCode accordingly.
+      //
+      // For more on this, read the article by Raymond Chen, titled:
+      // "Why does Ctrl+ScrollLock cancel dialogs?"
+      // https://devblogs.microsoft.com/oldnewthing/20080211-00/?p=23503
+      code = KeyCode::NumLock;
+    } else {
+      code = KeyCode::from_scancode(scancode as u32);
+    }
+    if keyboard.VKey as c_int == winuser::VK_SHIFT {
+      match code {
+        KeyCode::NumpadDecimal
+        | KeyCode::Numpad0
+        | KeyCode::Numpad1
+        | KeyCode::Numpad2
+        | KeyCode::Numpad3
+        | KeyCode::Numpad4
+        | KeyCode::Numpad5
+        | KeyCode::Numpad6
+        | KeyCode::Numpad7
+        | KeyCode::Numpad8
+        | KeyCode::Numpad9 => {
+          // On Windows, holding the Shift key makes numpad keys behave as if NumLock
+          // wasn't active. The way this is exposed to applications by the system is that
+          // the application receives a fake key release event for the shift key at the
+          // moment when the numpad key is pressed, just before receiving the numpad key
+          // as well.
+          //
+          // The issue is that in the raw device event (here), the fake shift release
+          // event reports the numpad key as the scancode. Unfortunately, the event doesn't
+          // have any information to tell whether it's the left shift or the right shift
+          // that needs to get the fake release (or press) event so we don't forward this
+          // event to the application at all.
+          //
+          // For more on this, read the article by Raymond Chen, titled:
+          // "The shift key overrides NumLock"
+          // https://devblogs.microsoft.com/oldnewthing/20040906-00/?p=37953
+          return;
+        }
+        _ => (),
+      }
+    }
+    subclass_input.send_event(Event::DeviceEvent {
+      device_id,
+      event: Key(RawKeyEvent {
+        physical_key: code,
+        state,
+      }),
+    });
+  }
+}
