@@ -3,6 +3,7 @@
 
 #![cfg(target_os = "windows")]
 
+use mem::MaybeUninit;
 use parking_lot::Mutex;
 use raw_window_handle::{windows::WindowsHandle, RawWindowHandle};
 use std::{
@@ -18,27 +19,29 @@ use winapi::{
   ctypes::c_int,
   shared::{
     minwindef::{HINSTANCE, LPARAM, UINT, WPARAM},
-    windef::{HWND, POINT, POINTS, RECT},
+    windef::{self, HWND, POINT, POINTS, RECT},
   },
   um::{
-    combaseapi, dwmapi,
+    combaseapi::{self, CoCreateInstance, CLSCTX_SERVER},
+    dwmapi,
     imm::{CFS_POINT, COMPOSITIONFORM},
     libloaderapi,
     objbase::COINIT_APARTMENTTHREADED,
     ole2,
     oleidl::LPDROPTARGET,
-    shobjidl_core::{CLSID_TaskbarList, ITaskbarList2},
+    shobjidl_core::{CLSID_TaskbarList, ITaskbarList, ITaskbarList2},
     wingdi::{CreateRectRgn, DeleteObject},
     winnt::{LPCWSTR, SHORT},
     winuser,
   },
+  Interface,
 };
 
 use crate::{
   dpi::{PhysicalPosition, PhysicalSize, Position, Size},
   error::{ExternalError, NotSupportedError, OsError as RootOsError},
   icon::Icon,
-  menu::Menu,
+  menu::MenuType,
   monitor::MonitorHandle as RootMonitorHandle,
   platform_impl::platform::{
     dark_mode::try_theme,
@@ -48,11 +51,15 @@ use crate::{
     icon::{self, IconType},
     menu, monitor, util,
     window_state::{CursorFlags, SavedWindow, WindowFlags, WindowState},
-    Parent, PlatformSpecificWindowBuilderAttributes, WindowId,
+    OsError, Parent, PlatformSpecificWindowBuilderAttributes, WindowId,
   },
   window::{CursorIcon, Fullscreen, Theme, UserAttentionType, WindowAttributes},
 };
 
+struct HMenuWrapper(windef::HMENU);
+unsafe impl Send for HMenuWrapper {}
+unsafe impl Sync for HMenuWrapper {}
+
 /// The Win32 implementation of the main `Window` object.
 pub struct Window {
   /// Main handle for the window.
@@ -63,6 +70,9 @@ pub struct Window {
 
   // The events loop proxy.
   thread_executor: event_loop::EventLoopThreadExecutor,
+
+  // The menu associated with the window
+  menu: Option<HMenuWrapper>,
 }
 
 impl Window {
@@ -139,7 +149,7 @@ impl Window {
   }
 
   // TODO (lemarier): allow menu update
-  pub fn set_menu(&self, _new_menu: Option<Vec<Menu>>) {}
+  pub fn set_menu(&self, _new_menu: Option<menu::Menu>) {}
 
   #[inline]
   pub fn set_visible(&self, visible: bool) {
@@ -154,8 +164,15 @@ impl Window {
 
   #[inline]
   pub fn set_focus(&self) {
-    unsafe {
-      force_window_active(self.window.0);
+    let window = self.window.clone();
+    let window_flags = self.window_state.lock().window_flags();
+
+    let is_visible = window_flags.contains(WindowFlags::VISIBLE);
+    let is_minimized = window_flags.contains(WindowFlags::MINIMIZED);
+    let is_foreground = window.0 == unsafe { winuser::GetForegroundWindow() };
+
+    if is_visible && !is_minimized && !is_foreground {
+      unsafe { force_window_active(window.0) };
     }
   }
 
@@ -325,7 +342,7 @@ impl Window {
         .lock()
         .mouse
         .set_cursor_flags(window.0, |f| f.set(CursorFlags::GRABBED, grab))
-        .map_err(|e| ExternalError::Os(os_error!(e)));
+        .map_err(|e| ExternalError::Os(os_error!(OsError::IoError(e))));
       let _ = tx.send(result);
     });
     rx.recv().unwrap()
@@ -361,10 +378,14 @@ impl Window {
     let mut point = POINT { x, y };
     unsafe {
       if winuser::ClientToScreen(self.window.0, &mut point) == 0 {
-        return Err(ExternalError::Os(os_error!(io::Error::last_os_error())));
+        return Err(ExternalError::Os(os_error!(OsError::IoError(
+          io::Error::last_os_error()
+        ))));
       }
       if winuser::SetCursorPos(point.x, point.y) == 0 {
-        return Err(ExternalError::Os(os_error!(io::Error::last_os_error())));
+        return Err(ExternalError::Os(os_error!(OsError::IoError(
+          io::Error::last_os_error()
+        ))));
       }
     }
     Ok(())
@@ -435,11 +456,17 @@ impl Window {
     window_state.window_flags.contains(WindowFlags::RESIZABLE)
   }
 
+  #[inline]
   pub fn is_decorated(&self) -> bool {
     let window_state = self.window_state.lock();
     window_state.window_flags.contains(WindowFlags::DECORATIONS)
   }
 
+  #[inline]
+  pub fn is_visible(&self) -> bool {
+    util::is_visible(self.window.0)
+  }
+
   #[inline]
   pub fn fullscreen(&self) -> Option<Fullscreen> {
     let window_state = self.window_state.lock();
@@ -703,6 +730,62 @@ impl Window {
   pub fn theme(&self) -> Theme {
     self.window_state.lock().current_theme
   }
+
+  #[inline]
+  pub fn set_skip_taskbar(&self, skip: bool) {
+    unsafe {
+      let mut taskbar_list: *mut ITaskbarList = std::mem::zeroed();
+      CoCreateInstance(
+        &CLSID_TaskbarList,
+        ptr::null_mut(),
+        CLSCTX_SERVER,
+        &ITaskbarList::uuidof(),
+        &mut taskbar_list as *mut _ as *mut _,
+      );
+      if skip {
+        (*taskbar_list).DeleteTab(self.hwnd());
+      } else {
+        (*taskbar_list).AddTab(self.hwnd());
+      }
+      (*taskbar_list).Release();
+    }
+  }
+
+  #[inline]
+  pub fn hide_menu(&self) {
+    unsafe {
+      winuser::SetMenu(self.hwnd(), ptr::null::<windef::HMENU>() as _);
+    }
+  }
+
+  #[inline]
+  pub fn show_menu(&self) {
+    if let Some(menu) = &self.menu {
+      unsafe {
+        winuser::SetMenu(self.hwnd(), menu.0);
+      }
+    }
+  }
+
+  #[inline]
+  pub fn reset_dead_keys(&self) {
+    // `ToUnicode` consumes the dead-key by default, so we are constructing a fake (but valid)
+    // key input which we can call `ToUnicode` with.
+    unsafe {
+      let vk = winuser::VK_SPACE as u32;
+      let scancode = winuser::MapVirtualKeyW(vk, winuser::MAPVK_VK_TO_VSC);
+      let kbd_state = [0; 256];
+      let mut char_buff = [MaybeUninit::uninit(); 8];
+      winuser::ToUnicode(
+        vk,
+        scancode,
+        kbd_state.as_ptr(),
+        char_buff[0].as_mut_ptr(),
+        char_buff.len() as i32,
+        0,
+      );
+    }
+  }
 }
 
 impl Drop for Window {
@@ -789,7 +872,7 @@ unsafe fn init<T: 'static>(
     );
 
     if handle.is_null() {
-      return Err(os_error!(io::Error::last_os_error()));
+      return Err(os_error!(OsError::IoError(io::Error::last_os_error())));
     }
 
     WindowWrapper(handle)
@@ -840,10 +923,11 @@ unsafe fn init<T: 'static>(
     window_state
   };
 
-  let win = Window {
+  let mut win = Window {
     window: real_window,
     window_state,
     thread_executor: event_loop.create_thread_executor(),
+    menu: None,
   };
 
   let dimensions = attributes
@@ -870,19 +954,24 @@ unsafe fn init<T: 'static>(
     let event_loop_runner = event_loop.runner_shared.clone();
     let window_handle = win.raw_window_handle();
 
-    let menu_handler = menu::MenuHandler::new(Box::new(move |event| {
-      if let Ok(e) = event.map_nonuser_event() {
-        event_loop_runner.send_event(e)
-      }
-    }));
+    let menu_handler = menu::MenuHandler::new(
+      Box::new(move |event| {
+        if let Ok(e) = event.map_nonuser_event() {
+          event_loop_runner.send_event(e)
+        }
+      }),
+      MenuType::MenuBar,
+    );
 
-    menu::initialize(window_menu, window_handle, menu_handler);
+    win.menu = menu::initialize(window_menu, window_handle, menu_handler).map(|m| HMenuWrapper(m));
   }
 
   if attributes.focus {
     force_window_active(win.window.0);
   }
 
+  win.set_skip_taskbar(attributes.skip_taskbar);
+
   Ok(win)
 }
 
@@ -965,7 +1054,7 @@ unsafe fn taskbar_mark_fullscreen(handle: HWND, fullscreen: bool) {
     let mut task_bar_list = task_bar_list_ptr.get();
 
     if task_bar_list == ptr::null_mut() {
-      use winapi::{shared::winerror::S_OK, Interface};
+      use winapi::shared::winerror::S_OK;
 
       let hr = combaseapi::CoCreateInstance(
         &CLSID_TaskbarList,
