@@ -8,7 +8,7 @@ use crate::{
   dpi::{PhysicalPosition, PhysicalSize, Position, Size},
   error::{ExternalError, NotSupportedError, OsError},
   event_loop::EventLoopWindowTarget,
-  menu::Menu,
+  menu::MenuBar,
   monitor::{MonitorHandle, VideoMode},
   platform_impl,
 };
@@ -201,7 +201,12 @@ pub struct WindowAttributes {
   /// The window menu.
   ///
   /// The default is `None`.
-  pub window_menu: Option<Vec<Menu>>,
+  pub window_menu: Option<platform_impl::Menu>,
+
+  /// Whether or not the window icon should be added to the taskbar.
+  ///
+  /// The default is `false`.
+  pub skip_taskbar: bool,
 }
 
 impl Default for WindowAttributes {
@@ -223,6 +228,7 @@ impl Default for WindowAttributes {
       always_on_top: false,
       window_icon: None,
       window_menu: None,
+      skip_taskbar: false,
     }
   }
 }
@@ -306,8 +312,8 @@ impl WindowBuilder {
   ///
   /// [`Window::set_menu`]: crate::window::Window::set_menu
   #[inline]
-  pub fn with_menu<M: Into<Vec<Menu>>>(mut self, menu: M) -> Self {
-    self.window.window_menu = Some(menu.into());
+  pub fn with_menu(mut self, menu: MenuBar) -> Self {
+    self.window.window_menu = Some(menu.0.menu_platform);
     self
   }
 
@@ -396,6 +402,17 @@ impl WindowBuilder {
     self
   }
 
+  /// Sets whether or not the window icon should be added to the taskbar.
+  ///
+  /// See [`Window::set_skip_taskbar`] for details.
+  ///
+  /// [`Window::set_skip_taskbar`]: crate::window::Window::set_skip_taskbar
+  #[inline]
+  pub fn with_skip_taskbar(mut self, skip: bool) -> Self {
+    self.window.skip_taskbar = skip;
+    self
+  }
+
   /// Builds the window.
   ///
   /// Possible causes of error include denied permission, incompatible system, and lack of memory.
@@ -613,8 +630,12 @@ impl Window {
   /// - **Windows:** Unsupported.
 
   #[inline]
-  pub fn set_menu(&self, menu: Option<Vec<Menu>>) {
-    self.window.set_menu(menu)
+  pub fn set_menu(&self, menu: Option<MenuBar>) {
+    if let Some(menu) = menu {
+      self.window.set_menu(Some(menu.0.menu_platform))
+    } else {
+      self.window.set_menu(None)
+    }
   }
 
   /// Modifies the window's visibility.
@@ -688,6 +709,16 @@ impl Window {
     self.window.is_maximized()
   }
 
+  /// Gets the window's current vibility state.
+  ///
+  /// ## Platform-specific
+  ///
+  /// - **iOS / Android:** Unsupported.
+  #[inline]
+  pub fn is_visible(&self) -> bool {
+    self.window.is_visible()
+  }
+
   /// Gets the window's current resizable state.
   ///
   /// ## Platform-specific
@@ -803,6 +834,40 @@ impl Window {
   pub fn request_user_attention(&self, request_type: Option<UserAttentionType>) {
     self.window.request_user_attention(request_type)
   }
+
+  /// Hides the menu associated with the window
+  ///
+  /// ## Platform-specific
+  ///
+  /// - **macOs/ iOS / Android:** Unsupported.
+  #[inline]
+  pub fn hide_menu(&self) {
+    self.window.hide_menu();
+  }
+
+  /// Shows the menu associated with the window
+  ///
+  /// ## Platform-specific
+  ///
+  /// - **macOs/ iOS / Android:** Unsupported.
+  #[inline]
+  pub fn show_menu(&self) {
+    self.window.show_menu();
+  }
+
+  /// Whether to show the window icon in the task bar or not.
+  ///
+  /// ## Platform-specific
+  ///
+  /// - **iOS / Android:** Unsupported.
+  ///
+  /// On macOS, you need to change the activation policy with
+  /// `event_loop.set_activation_policy(ActivationPolicy::Accessory);`
+  /// The `set_skip_taskbar` have no effect.
+  ///
+  pub fn set_skip_taskbar(&self, skip: bool) {
+    self.window.set_skip_taskbar(skip);
+  }
 }
 
 /// Cursor functions.
